/*
 * Copyright (c) 2022 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

// Package cacrt encapsulates the management of a CA certificate bundle
// -- a file in which the certificates in PEM format are concatenated.
package cacrt

import (
	"bytes"
	"crypto/sha512"
	"errors"
	"fmt"
	"io/ioutil"
	"os"
	"path/filepath"
	"strconv"
	"strings"
	"sync"
	"time"

	"code.uplex.de/k8s/k8s-crt-dnldr/pkg/crt"

	"golang.org/x/sys/unix"
)

const baseName = "ca-bundle.crt"

// A BundleSource specifies a k8s object from which content for a CA
// certificate is obtained. Currently we only support the type
// v1/Secret, i.e. APIVersion "v1" and Kind "Secret".
type BundleSource struct {
	APIVersion string `json:"apiVersion"`
	Kind       string `json:"kind"`
	Namespace  string `json:"namespace"`
	Name       string `json:"name"`
	UID        string `json:"uid,omitempty"`
	Version    string `json:"version,omitempty"`
	Key        string `json:"key,omitempty"`
}

// BundleSources specify all the content of a CA certificate bundle.
type BundleSources struct {
	// Srcs specify k8s objects from which content is obtained.
	Srcs []BundleSource `json:"sources,omitempty"`
	// Distro is true if the bundle includes a pre-installed CA
	// certificate bundle from the base image distribution.
	Distro bool `json:"default,omitempty"`
}

// File encapsulates the metadata of a CA certificate bundle -- the
// k8s Service for which it is intended, and file metadata.
type File struct {
	Sources BundleSources
	Size    int64
	ModTime time.Time
	Hash    [sha512.Size]byte
}

func (f File) String() string {
	return baseName
}

// Basename returns the file basename.
func (f File) Basename() string {
	return f.String()
}

// HaveSecret returns true iff the File includes the Secret identified
// by the given namespace, name, UID and ResourceVersion.
func (f File) HaveSecret(ns, name, uid, version string) bool {
	for _, src := range f.Sources.Srcs {
		if src.Namespace == ns && src.Name == name && src.UID == uid &&
			src.Version == version {
			return true
		}
	}
	return false
}

// Bundle encapsulates metadata about all of the CA bundles that are
// currently stored on the system, and provides capabilites to manage
// the files.
//
// Base is the name of the base directory in which bundles are stored.
//
// If Gid != -1, then it is a group ID set in the permissions of
// bundle files.
//
// Operations that read contents of the metadata and file system are
// protected by a shared mutex (read lock). Operations that change
// contents of the metadata or file system are protected by an
// exclusive lock (write lock).
type Bundle struct {
	File           *File
	Base           string
	Gid            int
	distroBndlPath string
	mtime          time.Time
	mtx            *sync.RWMutex
	crtGetter      *crt.Getter
}

// NewBundle returns a Bundle instance.
//
// base MUST be set to the name of a directory that is writable by the
// current process owner.
//
// If distroBndl is non-empty, it contains the path of a pre-installed
// CA certificate bundle stored in the base image. If a distroBndl is
// specified, it MUST identify a regular file that is readable by the
// current process owner.
//
// If gid is >= 0, then it is the group id set in bundle file
// permissions.
//
// crtGetter MUST be non-nil.
func NewBundle(
	base, distroBndl string, gid int, crtGetter *crt.Getter,
) (*Bundle, error) {
	if crtGetter == nil {
		return nil, errors.New("crtGetter is nil")
	}

	info, err := os.Stat(base)
	if err != nil {
		return nil, fmt.Errorf("Cannot stat %s: %v", base, err)
	}
	if !info.IsDir() {
		return nil, fmt.Errorf("%s: not a directory", base)
	}
	if err = unix.Access(base, unix.W_OK); err != nil {
		return nil, fmt.Errorf("%s not writable: %v", base, err)
	}
	if distroBndl != "" {
		info, err := os.Stat(distroBndl)
		if err != nil {
			return nil, fmt.Errorf("Cannot stat %s: %v", distroBndl,
				err)
		}
		if !info.Mode().IsRegular() {
			return nil, fmt.Errorf("%s: not a regular file",
				distroBndl)
		}
		if err = unix.Access(distroBndl, unix.R_OK); err != nil {
			return nil, fmt.Errorf("%s not readable: %v", base, err)
		}
	}

	bndl := &Bundle{
		Base:           base,
		Gid:            -1,
		distroBndlPath: distroBndl,
		File:           nil,
		mtx:            new(sync.RWMutex),
		mtime:          time.Now().UTC(),
		crtGetter:      crtGetter,
	}
	if gid >= 0 {
		bndl.Gid = gid
	}
	return bndl, nil
}

func (bndl *Bundle) String() string {
	s := make([]string, len(bndl.File.Sources.Srcs)+2)
	s[0] = "base:" + bndl.Base
	s[1] = "gid:" + strconv.Itoa(bndl.Gid)
	for i, src := range bndl.File.Sources.Srcs {
		s[i+2] = src.Namespace + "/" + src.Name
	}
	return "{" + strings.Join(s, ",") + "}"
}

// Path returns the file path to be used for a bundle file.
func (bndl *Bundle) Path(f *File) string {
	return filepath.Join(bndl.Base, f.Basename())
}

// ModTime returns the modification time of bundle metadata -- the
// time of the most recent Write(), AddOrUpdate, Delete() or
// DeleteSecret() call that resulted in a mutation of the bundle
// file; or of initalization of the metadata, if no modifications
// have taken place since NewBundle() was called.
func (bndl *Bundle) ModTime() time.Time {
	bndl.mtx.RLock()
	defer bndl.mtx.RUnlock()
	return bndl.mtime
}

// Have returns true iff the CA certificate bundle file exists
// according to the metadata.
//
// Have is protected by a read lock.
func (bndl *Bundle) Have() bool {
	bndl.mtx.RLock()
	defer bndl.mtx.RUnlock()

	return bndl.File != nil
}

// HaveSecret returns true iff one of the bundle Files includes the
// Secret specified by the given namespace, name, UID and
// ResourceVersion.
func (bndl *Bundle) HaveSecret(ns, name, uid, version string) bool {
	bndl.mtx.RLock()
	defer bndl.mtx.RUnlock()

	if bndl.File == nil {
		return false
	}
	return bndl.File.HaveSecret(ns, name, uid, version)
}

// HasDistroBundle returns true iff a pre-installed CA certificate
// bundle from the base image is configured for use.
func (bndl *Bundle) HasDistroBundle() bool {
	return bndl.distroBndlPath != ""
}

// Call under lock
func (bndl *Bundle) data(f *File) ([]byte, error) {
	buf := new(bytes.Buffer)
	for _, src := range f.Sources.Srcs {
		val, _, err := bndl.crtGetter.Find(
			src.Namespace, src.Name, src.UID, src.Version, src.Key)
		if err != nil {
			return nil, err
		}
		buf.Write(val)
	}
	if f.Sources.Distro {
		if bndl.distroBndlPath == "" {
			return nil, errors.New("Distro bundle not defined")
		}
		d, err := os.Open(bndl.distroBndlPath)
		if err != nil {
			return nil, err
		}
		defer d.Close()
		_, err = buf.ReadFrom(d)
		if err != nil {
			return nil, err
		}
	}
	return buf.Bytes(), nil
}

// Call under write lock
func (bndl *Bundle) write(f *File) error {
	data, err := bndl.data(f)
	if err != nil {
		return err
	}

	path := bndl.Path(f)
	if err = ioutil.WriteFile(path, data, os.FileMode(0640)); err != nil {
		return err
	}
	if bndl.Gid >= 0 {
		if err := os.Chown(path, -1, bndl.Gid); err != nil {
			return err
		}
	}
	info, err := os.Stat(path)
	if err != nil {
		return err
	}
	if info.Size() != int64(len(data)) {
		return fmt.Errorf("File written with unexpected size: "+
			"expected %d got %d", len(data), info.Size())
	}

	f.Hash = sha512.Sum512(data)
	f.Size = int64(len(data))
	f.ModTime = info.ModTime()
	bndl.mtime = time.Now().UTC()
	return nil
}

// call under lock
func (bndl *Bundle) checkFileData(f *File, hash [sha512.Size]byte) bool {
	path := bndl.Path(f)
	if data, err := bndl.data(f); err != nil {
		return false
	} else if sha512.Sum512(data) != hash {
		return false
	}
	if contents, err := ioutil.ReadFile(path); err != nil {
		return false
	} else if sha512.Sum512(contents) != hash {
		return false
	}
	return true
}

// call under lock
func (bndl *Bundle) check() bool {
	f := bndl.File
	if f == nil {
		return false
	}
	path := bndl.Path(f)
	info, err := os.Stat(path)
	if err != nil {
		return false
	}
	if info.Size() != f.Size {
		return false
	}
	if info.ModTime() != f.ModTime {
		return false
	}
	if info.Mode().Perm()&040 == 0 {
		return false
	}
	if bndl.Gid >= 0 {
		ustat := &unix.Stat_t{}
		if err = unix.Stat(path, ustat); err != nil {
			return false
		} else if int(ustat.Gid) != bndl.Gid {
			return false
		}
	}
	return bndl.checkFileData(f, f.Hash)
}

// Check returns true iff the current metadata and file contents for the
// CA bundle match the current configuration. In particular, Check
// returns false if a k8s object specified by the bundle has been
// changed, but the file contents have not yet been updated.
//
// false is returned if the hash value saved in the metadata does not
// match the hash of the current file contents, or if file metadata do
// not match the file currently stored on the system -- size and
// modification time.
//
// Moreover, the file must be group-readable, and if bndl.Gid != -1,
// then it must match the group ID in the files permissions.
//
// If all of these data match, then true is returned.
//
// Check is protected by a read lock.
func (bndl *Bundle) Check() bool {
	bndl.mtx.RLock()
	defer bndl.mtx.RUnlock()
	return bndl.check()
}

// CheckFileData returns true iff the contents specified by the given
// File object match the current contents stored in the bundle file.
//
// CheckFileData only checks the current file contents. It does not
// attempt to match the metadata stored in bndl, as the Check method
// does.
//
// CheckFileData is protected by a read lock.
func (bndl *Bundle) CheckFileData(f *File) bool {
	bndl.mtx.RLock()
	defer bndl.mtx.RUnlock()

	if bndl.File == nil {
		return false
	}
	return bndl.checkFileData(f, bndl.File.Hash)
}

// Write updates the contents of the CA bundle file according to current
// metadata.
//
// The write operation is successful if the return value is
// non-nil. The resulting file is group-readable but not accessible to
// other (file mode 0640). If bndl.Gid != -1, then the group ID in
// the file's permissions is set accordingly.
//
// This operation overwrites an existing file. To add a new bundle file,
// use the AddOrUpdate method.
//
// Write is protected by a write lock.
func (bndl *Bundle) Write() error {
	bndl.mtx.Lock()
	defer bndl.mtx.Unlock()

	if bndl.File == nil {
		return errors.New("Bundle.Write(): no file has been set")
	}
	return bndl.write(bndl.File)
}

// AddOrUpdate creates or updates the CA certificate bundle file
// specified by the given File. The metadata as well as file contents
// are added or modified.
//
// The file is named according to Basename(), and stored in the path
// set by the base parameter of NewBundle().
//
// The return value is non-nil if file is nil, or if there is a file
// I/O error.
//
// AddOrUpdate is protected by a write lock.
func (bndl *Bundle) AddOrUpdate(file *File) error {
	bndl.mtx.Lock()
	defer bndl.mtx.Unlock()

	if file == nil {
		return errors.New("File is nil")
	}

	if err := bndl.write(file); err != nil {
		return err
	}
	bndl.File = file
	return nil
}

// Delete deletes the bundle file, if the file exists.
//
// The boolean return value is true iff metadata was previously stored
// for the bundle, and the file was stored on the system before the
// attempted delete.
//
// The error return value is non-nil for any error deleting the file.
// If os.IsNotExist() is true of a non-nil error, then the boolean
// return value is false.
//
// Delete is protected by a write lock.
func (bndl *Bundle) Delete() (bool, error) {
	bndl.mtx.Lock()
	defer bndl.mtx.Unlock()

	if bndl.File == nil {
		return false, nil
	}

	path := bndl.Path(bndl.File)
	if err := os.Remove(path); err != nil {
		return !os.IsNotExist(err), err
	}
	bndl.File = nil
	bndl.mtime = time.Now().UTC()
	return true, nil
}

// DeleteSecret deletes the Secret specified by secretNS/secretName in
// the CA certificate bundle. The metadata stored in bndl as well as
// the file contents are updated.
//
// The return value is non-nil if the Secret is not included in the
// bundle file for the Service, or if there is a file I/O error.
//
// DeleteSecret is protected by a write lock.
func (bndl *Bundle) DeleteSecret(secretNS, secretName string) error {
	bndl.mtx.Lock()
	defer bndl.mtx.Unlock()

	file := bndl.File
	idx := -1
	for i, src := range file.Sources.Srcs {
		if src.Kind == "Secret" && src.Namespace == secretNS &&
			src.Name == secretName {
			idx = i
			break
		}
	}
	if idx == -1 {
		return fmt.Errorf("Secret %s/%s not in bundle metadata",
			secretNS, secretName)
	}

	// Delete from slice by moving the last item into the deleted
	// item's position, and truncating the slice.
	file.Sources.Srcs[idx] = file.Sources.Srcs[len(file.Sources.Srcs)-1]
	file.Sources.Srcs[len(file.Sources.Srcs)-1] = BundleSource{}
	file.Sources.Srcs = file.Sources.Srcs[:len(file.Sources.Srcs)-1]

	return bndl.write(file)
}

// GetFileInfo returns information suitable for JSON marshaling about
// a CA bundle file, if a file is currently stored for the given
// namespace and name.
//
// ns and name MUST be set, while uid and/or version MAY be empty. If
// either of uid or version is set, then the FileInfo is returned only
// if the UID and/or ResourceVersion of the corresponding Secret
// match.
//
// The found return value is true iff the corresponding PEM file is
// currently stored.
//
// If an entry is found, the PEM is read and its contents are parsed
// to populate the Cert field, to extent supported by go's crypto/x509
// facilities.
//
// GetFileInfo is protected by a read lock.
// func (bndl *Bundle) GetFileInfo(
// 	ns, name, uid, version string,
// ) (info FileInfo, found bool, err error) {
// 	mapKey := ns + "/" + name

// 	bndl.mtx.RLock()
// 	defer bndl.mtx.RUnlock()

// 	pemFile, exists := bndl.Files[mapKey]
// 	if !exists {
// 		err = fmt.Errorf("%s not found", mapKey)
// 		return
// 	}
// 	if uid != "" && pemFile.UID != uid {
// 		err = fmt.Errorf("%s uid=%s not found", mapKey, uid)
// 		return
// 	}
// 	if version != "" && pemFile.ResourceVersion != version {
// 		err = fmt.Errorf("%s version=%s not found", mapKey, version)
// 		return
// 	}

// 	path := bndl.Path(pemFile)
// 	pemBytes, err := ioutil.ReadFile(path)
// 	if err != nil {
// 		if !os.IsNotExist(err) {
// 			found = true
// 		}
// 		err = fmt.Errorf("%s: cannot read pem file: %v", mapKey, err)
// 		return
// 	}
// 	found = true

// 	crt, err := mkCrt(pemBytes)
// 	if err != nil {
// 		err = fmt.Errorf("%s: cannot read certificate: %v", mapKey, err)
// 		return
// 	}
// 	info.K8sSecret = Secret{
// 		Namespace: pemFile.Namespace,
// 		Name:      pemFile.Name,
// 		UID:       pemFile.UID,
// 		Version:   pemFile.ResourceVersion,
// 	}
// 	info.Meta = FileMeta{
// 		ModTime: pemFile.ModTime,
// 		Size:    pemFile.Size,
// 	}
// 	info.Cert = crt
// 	return
// }

// // GetAllFileInfo returns information about all of the PEM files
// // currently stored, suitable for JSON marshaling. Is is protected by
// // a read lock.
// func (bndl *Bundle) GetAllFileInfo() ([]FileInfo, error) {
// 	allInfo := make([]FileInfo, 0, len(bndl.Files))

// 	bndl.mtx.RLock()
// 	defer bndl.mtx.RUnlock()

// 	for _, pemFile := range bndl.Files {
// 		info, _, err := bndl.GetFileInfo(
// 			pemFile.Namespace, pemFile.Name, "", "")
// 		if err != nil {
// 			return allInfo, err
// 		}
// 		allInfo = append(allInfo, info)
// 	}
// 	return allInfo, nil
// }
