/*
 * Copyright (c) 2022 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package cacrt

import (
	"context"
	"crypto/sha512"
	"encoding/hex"
	"io"
	"io/ioutil"
	"os"
	"path/filepath"
	"testing"
	"time"

	"code.uplex.de/k8s/k8s-crt-dnldr/pkg/crt"

	api_v1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/client-go/informers"
	"k8s.io/client-go/kubernetes/fake"
	core_v1_listers "k8s.io/client-go/listers/core/v1"
	"k8s.io/client-go/tools/cache"
)

func setupSecretLister(client *fake.Clientset) core_v1_listers.SecretLister {
	ctx, cancel := context.WithCancel(context.Background())
	defer cancel()

	infFactory := informers.NewSharedInformerFactory(client, 0)
	informer := infFactory.Core().V1().Secrets().Informer()
	lister := infFactory.Core().V1().Secrets().Lister()

	infFactory.Start(ctx.Done())
	cache.WaitForCacheSync(ctx.Done(), informer.HasSynced)
	return lister
}

const expBaseName = "ca-bundle.crt"

func TestBasename(t *testing.T) {
	file := File{}
	if file.Basename() != expBaseName {
		t.Errorf("File.Basename(): got %s want %s", file.Basename(),
			expBaseName)
	}
}

func TestFileHaveSecret(t *testing.T) {
	file := File{
		Sources: BundleSources{
			Srcs: []BundleSource{
				{
					Namespace: "ns",
					Name:      "foo",
					UID:       "a9206402-f3c2-4ea8-9585-b0b44707b09b",
					Version:   "423267",
				},
				{
					Namespace: "ns",
					Name:      "bar",
					UID:       "fdc4d178-773a-4c1d-a145-0503dfcc819a",
					Version:   "365763",
				},
				{
					Namespace: "ns",
					Name:      "baz",
					UID:       "d381e245-f042-4049-9d8b-55aa4f68f3d7",
					Version:   "671207",
				},
			},
		},
	}
	if !file.HaveSecret("ns", "bar", "fdc4d178-773a-4c1d-a145-0503dfcc819a",
		"365763") {
		t.Error("file.HaveSecret(ns/bar): got false want true")
	}
	if file.HaveSecret("ns", "quux", "fdc4d178-773a-4c1d-a145-0503dfcc819a",
		"365763") {
		t.Error("file.HaveSecret(ns/quux): got true want false")
	}
	if file.HaveSecret("ns1", "bar", "fdc4d178-773a-4c1d-a145-0503dfcc819a",
		"365763") {
		t.Error("file.HaveSecret(ns1/bar): got true want false")
	}
	if file.HaveSecret("ns", "bar", "401aabcf-2781-4a84-a83c-91070d922441",
		"365763") {
		t.Error("file.HaveSecret(ns/bar, wrong UID): " +
			"got true want false")
	}
	if file.HaveSecret("ns", "bar", "fdc4d178-773a-4c1d-a145-0503dfcc819a",
		"585899") {
		t.Error("file.HaveSecret(ns/bar, wrong Version): " +
			"got false want true")
	}
}

func getTempDir() (string, error) {
	return ioutil.TempDir("", "k8s-crt-dnldr_cacrt_test")
}

func TestNewBundle(t *testing.T) {
	if _, err := NewBundle("", "", -1, nil); err == nil {
		t.Errorf("NewBundle(getter=nil) error: got nil want non-nil")
	}

	client := fake.NewSimpleClientset()
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)

	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	if _, err := NewBundle(base, "", -1, getter); err != nil {
		t.Errorf("NewBundle(%s) error: %+v", base, err)
	}

	noSuchDir := filepath.Join(base, "nosuchdir")
	if _, err := NewBundle(noSuchDir, "", -1, getter); err == nil {
		t.Errorf("NewBundle(no such base) error: got nil want non-nil")
	}
	nonDir := filepath.Join(base, "nondir")
	if err = ioutil.WriteFile(nonDir, []byte{}, os.FileMode(0664)); err != nil {
		t.Fatalf("Cannot write test file %s: %v", nonDir, err)
	}
	if _, err := NewBundle(nonDir, "", -1, getter); err == nil {
		t.Errorf("NewBundle(non dir base) error: got nil want non-nil")
	}

	if _, err := NewBundle(base, noSuchDir, -1, getter); err == nil {
		t.Errorf("NewBundle(no such distro bundle) error: got nil " +
			"want non-nil")
	}
	if _, err := NewBundle(base, base, -1, getter); err == nil {
		t.Errorf("NewBundle(distro bundle is dir) error: got nil " +
			"want non-nil")
	}
	if _, err := NewBundle(base, nonDir, -1, getter); err != nil {
		t.Errorf("NewBundle(distro bundle=%s) error: %+v", nonDir, err)
	}
}

func TestPath(t *testing.T) {
	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	client := fake.NewSimpleClientset()
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)

	bndl, err := NewBundle(base, "", -1, getter)
	if err != nil {
		t.Fatalf("NewBundle(): %v", err)
	}
	file := &File{}
	path := filepath.Join(base, expBaseName)
	if bndl.Path(file) != path {
		t.Errorf("Bundle.Path(): got %s want %s", bndl.Path(file), path)
	}
}

func TestHave(t *testing.T) {
	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	client := fake.NewSimpleClientset()
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)

	bndl, err := NewBundle(base, "", -1, getter)
	if err != nil {
		t.Fatalf("NewBundle(): %v", err)
	}
	if bndl.Have() {
		t.Error("Bundle.Have(): got true want false")
	}
	bndl.File = &File{}
	if !bndl.Have() {
		t.Error("Bundle.Have(): got false want true")
	}
}

func TestBundleHaveSecret(t *testing.T) {
	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	client := fake.NewSimpleClientset()
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)

	bndl, err := NewBundle(base, "", -1, getter)
	if err != nil {
		t.Fatalf("NewBundle(): %v", err)
	}

	bndl.File = &File{
		Sources: BundleSources{
			Srcs: []BundleSource{
				{
					Namespace: "ns",
					Name:      "foo",
					UID:       "a9206402-f3c2-4ea8-9585-b0b44707b09b",
					Version:   "423267",
				},
				{
					Namespace: "ns",
					Name:      "bar",
					UID:       "fdc4d178-773a-4c1d-a145-0503dfcc819a",
					Version:   "365763",
				},
				{
					Namespace: "ns",
					Name:      "baz",
					UID:       "d381e245-f042-4049-9d8b-55aa4f68f3d7",
					Version:   "671207",
				},
			},
		},
	}

	if !bndl.HaveSecret("ns", "bar", "fdc4d178-773a-4c1d-a145-0503dfcc819a",
		"365763") {
		t.Error("bndl.HaveSecret(ns/bar): got false want true")
	}
	if bndl.HaveSecret("ns", "quux", "fdc4d178-773a-4c1d-a145-0503dfcc819a",
		"365763") {
		t.Error("bndl.HaveSecret(ns/quux): got true want false")
	}
	if bndl.HaveSecret("ns1", "bar", "fdc4d178-773a-4c1d-a145-0503dfcc819a",
		"365763") {
		t.Error("bndl.HaveSecret(ns1/bar): got true want false")
	}
	if bndl.HaveSecret("ns", "bar", "401aabcf-2781-4a84-a83c-91070d922441",
		"365763") {
		t.Error("bndl.HaveSecret(ns/bar, wrong UID): " +
			"got true want false")
	}
	if bndl.HaveSecret("ns", "bar", "fdc4d178-773a-4c1d-a145-0503dfcc819a",
		"585899") {
		t.Error("bndl.HaveSecret(ns/bar, wrong Version): " +
			"got false want true")
	}
}

const (
	snakeOilFname = "foo_bar.crt"
	// from sha512sum
	snakeOilSha512Hex = "1fede4caa2d5430536dbb30c15bdcdf1d88bf8a15cbcf" +
		"d59cc0e890dbdb941592290aab4918977ae78f7f90f8bd00add2aa540110" +
		"76b80cf72f75001d723d126"
	snakeOilSz = 1099
)

func TestCheck(t *testing.T) {
	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	tpath := filepath.Join("testdata/", snakeOilFname)
	tfile, err := os.Open(tpath)
	if err != nil {
		t.Fatalf("os.Open(): %+v", err)
	}
	defer tfile.Close()
	snakeOil, err := io.ReadAll(tfile)
	if err != nil {
		t.Fatalf("io.ReadAll(%s): %+v", snakeOilFname, err)
	}

	client := fake.NewSimpleClientset(
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "foo",
				Name:            "bar",
				UID:             "21653844-1aa5-4182-8260-c077b69bcaba",
				ResourceVersion: "670288",
			},
			Data: map[string][]byte{
				"cacrt": snakeOil,
			},
		},
	)
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)

	fpath := filepath.Join(base, expBaseName)
	if err = os.WriteFile(fpath, snakeOil, 0640); err != nil {
		t.Fatalf("os.WriteFile(%s): %+v", fpath, err)
	}
	stat, err := os.Stat(fpath)
	if err != nil {
		t.Fatalf("os.Stat(): %+v", err)
	}

	hash, err := hex.DecodeString(snakeOilSha512Hex)
	if err != nil {
		t.Fatalf("hex.DecodeString(): %+v", err)
	}
	hashPtr := (*[sha512.Size]byte)(hash)

	bndl, err := NewBundle(base, "", -1, getter)
	if err != nil {
		t.Fatalf("NewBundle(): %v", err)
	}
	bndl.File = &File{
		Size:    snakeOilSz,
		Hash:    *hashPtr,
		ModTime: stat.ModTime(),
		Sources: BundleSources{
			Srcs: []BundleSource{
				{
					Namespace: "foo",
					Name:      "bar",
					UID:       "21653844-1aa5-4182-8260-c077b69bcaba",
					Version:   "670288",
					Key:       "cacrt",
				},
			},
		},
	}

	if !bndl.Check() {
		t.Error("Check() got false want true")
	}
	hashCpy := *hashPtr
	hashCpy[0] = byte(0)
	bndl.File.Hash = hashCpy
	if bndl.Check() {
		t.Error("Check(wrong hash) got true want false")
	}
	bndl.File.Size = snakeOilSz + 1
	if bndl.Check() {
		t.Error("Check(wrong size) got true want false")
	}
	bndl.File.Size = snakeOilSz
	bndl.File.ModTime = time.Now().Add(time.Second)
	if bndl.Check() {
		t.Error("Check(wrong mtime) got true want false")
	}
}

func TestCheckFileData(t *testing.T) {
	tpath := filepath.Join("testdata/", snakeOilFname)
	tfile, err := os.Open(tpath)
	if err != nil {
		t.Fatalf("os.Open(): %+v", err)
	}
	defer tfile.Close()
	snakeOil, err := io.ReadAll(tfile)
	if err != nil {
		t.Fatalf("io.ReadAll(%s): %+v", snakeOilFname, err)
	}

	client := fake.NewSimpleClientset(
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "foo",
				Name:            "bar",
				UID:             "7a85f291-1a2f-418a-a4f9-54e86ce24bfb",
				ResourceVersion: "875261",
			},
			Data: map[string][]byte{
				"cacrt": snakeOil,
			},
		},
	)
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)

	hash, err := hex.DecodeString(snakeOilSha512Hex)
	if err != nil {
		t.Fatalf("hex.DecodeString(): %+v", err)
	}
	hashPtr := (*[sha512.Size]byte)(hash)

	bndl, err := NewBundle("testdata/", "", -1, getter)
	if err != nil {
		t.Fatalf("NewBundle(): %v", err)
	}
	file := &File{
		Hash: *hashPtr,
		Sources: BundleSources{
			Srcs: []BundleSource{
				{
					Namespace: "foo",
					Name:      "bar",
					UID:       "7a85f291-1a2f-418a-a4f9-54e86ce24bfb",
					Version:   "875261",
					Key:       "cacrt",
				},
			},
		},
	}
	bndl.File = file

	if !bndl.CheckFileData(file) {
		t.Error("Check() got false want true")
	}
	hashCpy := *hashPtr
	hashCpy[0] = byte(0)
	file.Hash = hashCpy
	if bndl.CheckFileData(file) {
		t.Error("Check(wrong hash) got true want false")
	}
}

var fakeCrt = []byte("fake CA certificate contents")

const fakeSha512Hex = "bde3ee0fd504bbb2aa7738948e48950e7e93722790a91e01846f6050adeba63b90682a324958189d7d5ef435228ddc679534ca1122b9a2f00820ab77587367d1"

func TestWrite(t *testing.T) {
	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	client := fake.NewSimpleClientset(
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "foo",
				Name:            "bar",
				UID:             "d8cb68e0-a004-4f2a-b16d-c79ba4edf4dd",
				ResourceVersion: "742737",
			},
			Data: map[string][]byte{
				"crtkey": fakeCrt,
			},
		},
	)
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)

	bndl, err := NewBundle(base, "", -1, getter)
	if err != nil {
		t.Fatalf("NewBundle(): %v", err)
	}
	file := &File{
		Sources: BundleSources{
			Srcs: []BundleSource{
				{
					Namespace: "foo",
					Name:      "bar",
					UID:       "d8cb68e0-a004-4f2a-b16d-c79ba4edf4dd",
					Version:   "742737",
					Key:       "crtkey",
				},
			},
		},
	}
	bndl.File = file

	if err = bndl.Write(); err != nil {
		t.Errorf("Bundle.Write(foo, bar): %+v", err)
	}

	hash, err := hex.DecodeString(fakeSha512Hex)
	if err != nil {
		t.Fatalf("hex.DecodeString(): %+v", err)
	}
	var h64 [64]byte
	copy(h64[:], hash)
	if h64 != file.Hash {
		t.Errorf("Hash after Bundle.Write(foo, bar): want %s got %s",
			snakeOilSha512Hex, file.Hash[:])
	}

	if int64(len(fakeCrt)) != file.Size {
		t.Errorf("Size after Bundle.Write(foo, bar): want %d got %d",
			len(fakeCrt), file.Size)
	}

	if !bndl.CheckFileData(file) {
		t.Error("Check() got false want true")
	}
	if !bndl.Check() {
		t.Error("Bundle.Check() after .Write() got false want true")
	}
}

func TestAddOrUpdate(t *testing.T) {
	gid := os.Getgid()
	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	client := fake.NewSimpleClientset(
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "namespace",
				Name:            "name",
				UID:             "73e880f3-d7ca-447f-b95e-88f220d4820e",
				ResourceVersion: "654321",
			},
			Data: map[string][]byte{
				"crt": fakeCrt,
			},
		},
	)
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)

	hash := sha512.Sum512(fakeCrt)

	bndl, err := NewBundle(base, "", gid, getter)
	if err != nil {
		t.Fatalf("NewBundle(): %v", err)
	}
	b4 := bndl.ModTime()

	bndl.File = nil
	file := &File{
		Sources: BundleSources{
			Srcs: []BundleSource{{
				APIVersion: "v1",
				Kind:       "Secret",
				Namespace:  "namespace",
				Name:       "name",
				UID:        "73e880f3-d7ca-447f-b95e-88f220d4820e",
				Version:    "654321",
				Key:        "crt",
			}},
		},
	}
	if err := bndl.AddOrUpdate(file); err != nil {
		t.Fatalf("AddOrUpdate: %+v", err)
	}

	f := bndl.File
	if f == nil {
		t.Fatal("Write(): File field not updated")
	}
	if f.Size != int64(len(fakeCrt)) {
		t.Errorf("Write(): File.Size want %d got %d", len(fakeCrt),
			f.Size)
	}
	if f.Hash != hash {
		t.Error("Write(): File.Hash incorrect value")
	}
	if !bndl.ModTime().After(b4) {
		t.Error("Write(): Bundle.ModTime not updated")
	}

	path := filepath.Join(base, expBaseName)
	info, err := os.Stat(path)
	if err != nil {
		t.Errorf("os.Stat(): %+v", err)
	}
	if info.Size() != f.Size {
		t.Errorf("os.Stat().Size()=%d, expected %d", info.Size(),
			f.Size)
	}
	if info.ModTime() != f.ModTime {
		statTime := info.ModTime().Format(time.RFC3339Nano)
		fTime := f.ModTime.Format(time.RFC3339Nano)
		t.Errorf("os.Stat().ModTime()=%s, expected %s", statTime, fTime)
	}

	contents, err := ioutil.ReadFile(path)
	if err != nil {
		t.Fatalf("ReadFile(): %+v", err)
	}
	if len(contents) != len(fakeCrt) {
		t.Errorf("Write() readback length: want %d got %d",
			len(fakeCrt), len(contents))
	}
	for i := range contents {
		if contents[i] != fakeCrt[i] {
			t.Errorf("Write() readback byte %d: want %d got %d",
				i, fakeCrt[i], contents[i])
		}
	}
	if sha512.Sum512(contents) != hash {
		t.Error("Write() readback: hash mismatch")
	}

	if !bndl.Have() {
		t.Error("Have(): want true got false")
	}
	if !bndl.Check() {
		t.Error("Check(): want true got false")
	}
	if !bndl.CheckFileData(file) {
		t.Error("CheckFileData(): want true got false")
	}
}

func TestDelete(t *testing.T) {
	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	testPath := filepath.Join("testdata", "foo_bar.crt")
	testFile, err := os.Open(testPath)
	if err != nil {
		t.Fatalf("os.Open(%s): %+v", testPath, err)
	}
	defer testFile.Close()
	crtPath := filepath.Join(base, expBaseName)
	crtFile, err := os.Create(crtPath)
	if err != nil {
		t.Fatalf("os.Create(%s): %+v", crtPath, err)
	}
	if n, err := io.Copy(crtFile, testFile); err != nil {
		t.Fatalf("io.Copy(%s, %s): %+v", crtPath, testPath, err)
	} else if n != snakeOilSz {
		t.Fatalf("io.Copy(%s, %s) bytes written: want %d got %d",
			crtPath, testPath, snakeOilSz, n)
	}
	crtFile.Close()

	client := fake.NewSimpleClientset()
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)

	bndl, err := NewBundle(base, "", -1, getter)
	if err != nil {
		t.Fatalf("NewBundle(): %v", err)
	}
	bndl.File = &File{}
	b4 := time.Now().UTC()

	if exists, err := bndl.Delete(); !exists {
		t.Fatal("Bundle.Delete() exists: want true got false")
	} else if err != nil {
		t.Fatalf("Bundle.Delete(): %+v", err)
	}

	if _, err := os.Stat(crtPath); err == nil {
		t.Error("os.Stat() err after Bundle.Delete(): " +
			"want non-nil got nil")
	}

	if bndl.File != nil {
		t.Error("Bundle.Delete(): file not deleted from metadata")
	}

	if !bndl.ModTime().After(b4) {
		t.Error("Bundle.Delete(): mtime not updated")
	}
}

func TestDeleteSecret(t *testing.T) {
	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	client := fake.NewSimpleClientset(
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "ns",
				Name:            "foo",
				UID:             "c3aebfbf-0a35-41dc-a565-6936a1335c84",
				ResourceVersion: "123440",
			},
			Data: map[string][]byte{
				"fookey": []byte("fake foo CA crt"),
			},
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "ns",
				Name:            "bar",
				UID:             "81bb0cf9-d7aa-4f12-b5f8-107441474d69",
				ResourceVersion: "144921",
			},
			Data: map[string][]byte{
				"barkey": []byte("fake bar CA crt"),
			},
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "ns",
				Name:            "baz",
				UID:             "cceff483-855d-4f3b-beba-2d37c8361889",
				ResourceVersion: "458449",
			},
			Data: map[string][]byte{
				"bazkey": []byte("fake baz CA crt"),
			},
		},
	)
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)

	file := &File{
		Sources: BundleSources{
			Srcs: []BundleSource{
				{
					Kind:      "Secret",
					Namespace: "ns",
					Name:      "foo",
					UID:       "c3aebfbf-0a35-41dc-a565-6936a1335c84",
					Version:   "123440",
					Key:       "fookey",
				},
				{
					Kind:      "Secret",
					Namespace: "ns",
					Name:      "bar",
					UID:       "81bb0cf9-d7aa-4f12-b5f8-107441474d69",
					Version:   "144921",
					Key:       "barkey",
				},
				{
					Kind:      "Secret",
					Namespace: "ns",
					Name:      "baz",
					UID:       "cceff483-855d-4f3b-beba-2d37c8361889",
					Version:   "458449",
					Key:       "bazkey",
				},
			},
		},
	}

	bndl, err := NewBundle(base, "", -1, getter)
	if err != nil {
		t.Fatalf("NewBundle(): %v", err)
	}

	if err := bndl.AddOrUpdate(file); err != nil {
		t.Fatalf("Bundle.AddOrUpdate(): %v", err)
	}
	b4 := bndl.ModTime()

	if err := bndl.DeleteSecret("ns", "bar"); err != nil {
		t.Fatalf("Bundle.Delete(): %+v", err)
	}

	if bndl.HaveSecret("ns", "bar", "81bb0cf9-d7aa-4f12-b5f8-107441474d69",
		"144921") {
		t.Error("Bundle.HaveSecret() after DeleteSecret(): " +
			"want false got true")
	}

	if !bndl.ModTime().After(b4) {
		t.Error("Bundle.DeleteSecret(): mtime not updated")
	}

	crtPath := filepath.Join(base, expBaseName)
	contents, err := ioutil.ReadFile(crtPath)
	if err != nil {
		t.Fatalf("iotuil.ReadFile(%s): %+v", crtPath, err)
	}
	if string(contents) != "fake foo CA crtfake baz CA crt" {
		t.Errorf("Contents after Bundle.DeleteSecret(): "+
			"want 'fake foo CA crtfake baz CA crt' got %s",
			contents)
	}
}
