/*
 * Copyright (c) 2020 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package rest

import (
	"bytes"
	_ "crypto/sha512"
	"encoding/json"
	"io"
	"io/ioutil"
	"net/http"
	"net/http/httptest"
	"os"
	"path/filepath"
	"strconv"
	"strings"
	"testing"

	"code.uplex.de/k8s/k8s-crt-dnldr/pkg/cacrt"
	"code.uplex.de/k8s/k8s-crt-dnldr/pkg/crt"
	"code.uplex.de/k8s/k8s-crt-dnldr/pkg/pem"

	"github.com/sirupsen/logrus"

	api_v1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/client-go/kubernetes/fake"
)

const reqPath = "/v1/ca-bndl"

func TestPutBndl(t *testing.T) {
	gid := os.Getgid()
	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)
	distroDir, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(distroDir)
	distroPath := filepath.Join(distroDir, "distroBndl.crt")
	if err = ioutil.WriteFile(distroPath, []byte("fake distro CA certs"),
		0640); err != nil {
		t.Fatalf("ioutil.WriteFile(distro): %+v", err)
	}

	client := fake.NewSimpleClientset(
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "ns1",
				Name:            "secret1",
				UID:             "2572e346-0710-4c3d-a1ae-bacaddb6f3f9",
				ResourceVersion: "4711",
			},
			Data: map[string][]byte{
				"cacrt": []byte("ca-crt1"),
			},
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "ns2",
				Name:            "secret2",
				UID:             "1ce39371-b2a4-43f7-9e1d-240db0b933a4",
				ResourceVersion: "0815",
			},
			Data: map[string][]byte{
				"ca-crt": []byte("ca-crt2"),
			},
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "ns3",
				Name:            "secret3",
				UID:             "204ebad6-e06d-4e94-a0c5-82d5968b4bc4",
				ResourceVersion: "123456",
			},
			Data: map[string][]byte{
				"cacert": []byte("CA certificate"),
			},
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "ns4",
				Name:            "secret4",
				UID:             "97b45482-bf35-4949-9319-fb2f9aa1eeac",
				ResourceVersion: "654321",
			},
			Data: map[string][]byte{
				"CAcert": []byte("certificate authority"),
			},
		},
	)
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)
	bndl, err := cacrt.NewBundle(base, distroPath, gid, getter)
	if err != nil {
		t.Fatalf("NewBundle(): %v", err)
	}

	hndlr := &bndlHndlr{
		log:  &logrus.Logger{Out: ioutil.Discard},
		bndl: bndl,
		crt:  getter,
	}

	srcs := cacrt.BundleSources{
		Srcs: []cacrt.BundleSource{{
			APIVersion: "v1",
			Kind:       "Secret",
			Namespace:  "ns1",
			Name:       "secret1",
			Key:        "cacrt",
		}},
		Distro: false,
	}
	srcsJSON, err := json.Marshal(srcs)
	if err != nil {
		t.Fatalf("json.Marshal(cacrt.BundleSources): %+v", err)
	}
	body := bytes.NewReader(srcsJSON)
	req := httptest.NewRequest(http.MethodPut, reqPath, body)
	req.Header.Set("Content-Type", "application/json")
	rr := httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusCreated {
		t.Errorf("PUT status: got %d want %d", rr.Code,
			http.StatusCreated)
		if testing.Verbose() {
			t.Logf("Body: %q", rr.Body.Bytes())
		}
	}
	expected := []byte("ca-crt1")
	if !bndl.Check() {
		t.Error("files.Check() after PUT: got false want true")
	}
	file := bndl.File
	if file == nil {
		t.Fatal("no file map entry after PUT")
	}
	bndlPath := bndl.Path(file)
	if contents, err := ioutil.ReadFile(bndlPath); err != nil {
		t.Fatalf("ioutil.ReadFile(%s) after PUT: %+v", bndlPath, err)
	} else if !bytes.Equal(contents, expected) {
		t.Errorf("file contents after PUT: got %q want %q", contents,
			expected)
	}
	if rr.Result().Header.Get("Content-Length") != "0" {
		t.Errorf("PUT /v1/ca-crt Content-Length: got %s want 0",
			rr.Result().Header.Get("Content-Length"))
	}

	// Rewind the bytes.Reader, so the request body can be read again.
	body.Seek(0, io.SeekStart)
	req = httptest.NewRequest(http.MethodPut, reqPath, body)
	req.Header.Set("Content-Type", "application/json")
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNoContent {
		t.Errorf("2nd PUT status: got %d want %d", rr.Code,
			http.StatusNoContent)
		if testing.Verbose() {
			t.Logf("Body: %q", rr.Body.Bytes())
		}
	}
	if rr.Result().Header.Get("Content-Length") != "" {
		t.Errorf("2nd PUT Content-Length: got %s want empty",
			rr.Result().Header.Get("Content-Length"))
	}

	// ns/name not found
	srcs = cacrt.BundleSources{
		Srcs: []cacrt.BundleSource{{
			APIVersion: "v1",
			Kind:       "Secret",
			Namespace:  "foo",
			Name:       "bar",
			Key:        "baz",
		}},
		Distro: false,
	}
	srcsJSON, err = json.Marshal(srcs)
	if err != nil {
		t.Fatalf("json.Marshal(cacrt.BundleSources): %+v", err)
	}
	body = bytes.NewReader(srcsJSON)
	req = httptest.NewRequest(http.MethodPut, reqPath, body)
	req.Header.Set("Content-Type", "application/json")
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNotFound {
		t.Errorf("PUT status: got %d want %d", rr.Code,
			http.StatusNotFound)
		if testing.Verbose() {
			t.Logf("Body: %q", rr.Body.Bytes())
		}
	}
	if rr.Header().Get("Content-Type") != problemContentType {
		t.Errorf("PUT Content-Type: got %s want %s",
			rr.Header().Get("Content-Type"), problemContentType)
	}
	bodylen := len(rr.Body.String())
	if rr.Header().Get("Content-Length") != strconv.Itoa(bodylen) {
		t.Errorf("PUT Content-Length: got %s want %d",
			rr.Header().Get("Content-Length"), bodylen)
	}
	problem := &Problem{}
	if err = json.Unmarshal(rr.Body.Bytes(), problem); err != nil {
		t.Fatalf("PUT body unmarshal: %v", err)
	}
	if problem.Type != errBndlNotFound.Type {
		t.Errorf("PUT problem type: got %s want %s", problem.Type,
			errBndlNotFound.Type)
	}
	if problem.Title != errBndlNotFound.Title {
		t.Errorf("PUT problem title: got %s want %s", problem.Title,
			errBndlNotFound.Title)
	}
	errMsg := "v1/Secret foo/bar key=baz not found"
	if !strings.HasPrefix(problem.Detail, errMsg) {
		t.Errorf("PUT problem detail: got \"%s\" want preifx \"%s\"",
			problem.Detail, errMsg)
	}
	if problem.Status != http.StatusNotFound {
		t.Errorf("PUT problem status: got %d want %d", problem.Status,
			http.StatusNotFound)
	}
	if !errInstancePattern.Match([]byte(problem.Instance)) {
		t.Errorf("PUT problem instance: got %s want /log/errors/N",
			problem.Instance)
	}

	// Key not found
	srcs = cacrt.BundleSources{
		Srcs: []cacrt.BundleSource{{
			APIVersion: "v1",
			Kind:       "Secret",
			Namespace:  "ns1",
			Name:       "secret1",
			Key:        "invalid",
		}},
		Distro: false,
	}
	srcsJSON, err = json.Marshal(srcs)
	if err != nil {
		t.Fatalf("json.Marshal(cacrt.BundleSources): %+v", err)
	}
	body = bytes.NewReader(srcsJSON)
	req = httptest.NewRequest(http.MethodPut, reqPath, body)
	req.Header.Set("Content-Type", "application/json")
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNotFound {
		t.Errorf("PUT status: got %d want %d", rr.Code,
			http.StatusNotFound)
		if testing.Verbose() {
			t.Logf("Body: %q", rr.Body.Bytes())
		}
	}
	if rr.Header().Get("Content-Type") != problemContentType {
		t.Errorf("PUT Content-Type: got %s want %s",
			rr.Header().Get("Content-Type"), problemContentType)
	}
	bodylen = len(rr.Body.String())
	if rr.Header().Get("Content-Length") != strconv.Itoa(bodylen) {
		t.Errorf("PUT Content-Length: got %s want %d",
			rr.Header().Get("Content-Length"), bodylen)
	}
	problem = &Problem{}
	if err = json.Unmarshal(rr.Body.Bytes(), problem); err != nil {
		t.Fatalf("PUT body unmarshal: %v", err)
	}
	if problem.Type != errBndlNotFound.Type {
		t.Errorf("PUT problem type: got %s want %s", problem.Type,
			errBndlNotFound.Type)
	}
	if problem.Title != errBndlNotFound.Title {
		t.Errorf("PUT problem title: got %s want %s", problem.Title,
			errBndlNotFound.Title)
	}
	errMsg = "v1/Secret ns1/secret1 key=invalid not found"
	if !strings.HasPrefix(problem.Detail, errMsg) {
		t.Errorf("PUT problem detail: got \"%s\" want preifx \"%s\"",
			problem.Detail, errMsg)
	}
	if problem.Status != http.StatusNotFound {
		t.Errorf("PUT problem status: got %d want %d", problem.Status,
			http.StatusNotFound)
	}
	if !errInstancePattern.Match([]byte(problem.Instance)) {
		t.Errorf("PUT problem instance: got %s want /log/errors/N",
			problem.Instance)
	}

	bndl.File = nil // Reset
	srcs = cacrt.BundleSources{
		Srcs: []cacrt.BundleSource{{
			APIVersion: "v1",
			Kind:       "Secret",
			Namespace:  "ns2",
			Name:       "secret2",
			UID:        "1ce39371-b2a4-43f7-9e1d-240db0b933a4",
			Version:    "0815",
			Key:        "ca-crt",
		}},
		Distro: false,
	}
	srcsJSON, err = json.Marshal(srcs)
	if err != nil {
		t.Fatalf("json.Marshal(cacrt.BundleSources): %+v", err)
	}
	body = bytes.NewReader(srcsJSON)
	req = httptest.NewRequest(http.MethodPut, reqPath, body)
	req.Header.Set("Content-Type", "application/json")
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusCreated {
		t.Errorf("PUT status: got %d want %d", rr.Code,
			http.StatusCreated)
		if testing.Verbose() {
			t.Logf("Body: %q", rr.Body.Bytes())
		}
	}
	expected = []byte("ca-crt2")
	if !bndl.Check() {
		t.Error("bndl.Check() after PUT: got false want true")
	}
	f := bndl.File
	if f == nil {
		t.Fatal("no file metadata after PUT")
	}
	bndlPath = bndl.Path(f)
	if contents, err := ioutil.ReadFile(bndlPath); err != nil {
		t.Fatalf("ioutil.ReadFile(%s) after PUT: %+v", bndlPath,
			err)
	} else if !bytes.Equal(contents, expected) {
		t.Errorf("file contents after PUT: got %q want %q", contents,
			expected)
	}

	bndl.File = nil // Reset
	srcs = cacrt.BundleSources{
		Srcs: []cacrt.BundleSource{{
			APIVersion: "v1",
			Kind:       "Secret",
			Namespace:  "ns3",
			Name:       "secret3",
			UID:        "204ebad6-e06d-4e94-a0c5-82d5968b4bc4",
			Key:        "cacert",
		}},
		Distro: false,
	}
	srcsJSON, err = json.Marshal(srcs)
	if err != nil {
		t.Fatalf("json.Marshal(cacrt.BundleSources): %+v", err)
	}
	body = bytes.NewReader(srcsJSON)
	req = httptest.NewRequest(http.MethodPut, reqPath, body)
	req.Header.Set("Content-Type", "application/json")
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusCreated {
		t.Errorf("PUT status: got %d want %d", rr.Code,
			http.StatusCreated)
		if testing.Verbose() {
			t.Logf("Body: %q", rr.Body.Bytes())
		}
	}
	expected = []byte("CA certificate")
	if !bndl.Check() {
		t.Error("files.Check() after PUT: got false want true")
	}
	f = bndl.File
	if f == nil {
		t.Fatal("no file metadata after PUT")
	}
	bndlPath = bndl.Path(f)
	if contents, err := ioutil.ReadFile(bndlPath); err != nil {
		t.Fatalf("ioutil.ReadFile(%s) after PUT: %+v", bndlPath,
			err)
	} else if !bytes.Equal(contents, expected) {
		t.Errorf("file contents after PUT: got %q want %q", contents,
			expected)
	}

	bndl.File = nil // Reset
	srcs = cacrt.BundleSources{
		Srcs: []cacrt.BundleSource{{
			APIVersion: "v1",
			Kind:       "Secret",
			Namespace:  "ns4",
			Name:       "secret4",
			Version:    "654321",
			Key:        "CAcert",
		}},
		Distro: false,
	}
	srcsJSON, err = json.Marshal(srcs)
	if err != nil {
		t.Fatalf("json.Marshal(cacrt.BundleSources): %+v", err)
	}
	body = bytes.NewReader(srcsJSON)
	req = httptest.NewRequest(http.MethodPut, reqPath, body)
	req.Header.Set("Content-Type", "application/json")
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusCreated {
		t.Errorf("PUT status: got %d want %d", rr.Code,
			http.StatusCreated)
		if testing.Verbose() {
			t.Logf("Body: %q", rr.Body.Bytes())
		}
	}
	expected = []byte("certificate authority")
	if !bndl.Check() {
		t.Error("bndl.Check() after PUT: got false want true")
	}
	f = bndl.File
	if f == nil {
		t.Fatal("no file metadata after PUT")
	}
	bndlPath = bndl.Path(f)
	if contents, err := ioutil.ReadFile(bndlPath); err != nil {
		t.Fatalf("ioutil.ReadFile(%s) after PUT: %+v", bndlPath, err)
	} else if !bytes.Equal(contents, expected) {
		t.Errorf("file contents after PUT: got %q want %q", contents,
			expected)
	}

	bndl.File = nil // Reset
	srcs = cacrt.BundleSources{
		Distro: true,
	}
	srcsJSON, err = json.Marshal(srcs)
	if err != nil {
		t.Fatalf("json.Marshal(cacrt.BundleSources): %+v", err)
	}
	body = bytes.NewReader(srcsJSON)
	req = httptest.NewRequest(http.MethodPut, reqPath, body)
	req.Header.Set("Content-Type", "application/json")
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusCreated {
		t.Errorf("PUT status: got %d want %d", rr.Code,
			http.StatusCreated)
		if testing.Verbose() {
			t.Logf("Body: %q", rr.Body.Bytes())
		}
	}
	expected = []byte("fake distro CA certs")
	if !bndl.Check() {
		t.Error("bndl.Check() after PUT: got false want true")
	}
	f = bndl.File
	if f == nil {
		t.Fatal("no file metadta after PUT")
	}
	bndlPath = bndl.Path(f)
	if contents, err := ioutil.ReadFile(bndlPath); err != nil {
		t.Fatalf("ioutil.ReadFile(%s) after PUT: %+v", bndlPath, err)
	} else if !bytes.Equal(contents, expected) {
		t.Errorf("file contents after PUT: got %q want %q", contents,
			expected)
	}

	bndl.File = nil // Reset
	srcs = cacrt.BundleSources{
		Srcs: []cacrt.BundleSource{
			{
				APIVersion: "v1",
				Kind:       "Secret",
				Namespace:  "ns1",
				Name:       "secret1",
				Key:        "cacrt",
			},
			{
				APIVersion: "v1",
				Kind:       "Secret",
				Namespace:  "ns2",
				Name:       "secret2",
				Key:        "ca-crt",
			},
			{
				APIVersion: "v1",
				Kind:       "Secret",
				Namespace:  "ns3",
				Name:       "secret3",
				Key:        "cacert",
			},
			{
				APIVersion: "v1",
				Kind:       "Secret",
				Namespace:  "ns4",
				Name:       "secret4",
				Key:        "CAcert",
			},
		},
		Distro: true,
	}
	srcsJSON, err = json.Marshal(srcs)
	if err != nil {
		t.Fatalf("json.Marshal(cacrt.BundleSources): %+v", err)
	}
	body = bytes.NewReader(srcsJSON)
	req = httptest.NewRequest(http.MethodPut, reqPath, body)
	req.Header.Set("Content-Type", "application/json")
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusCreated {
		t.Errorf("PUT status: got %d want %d", rr.Code,
			http.StatusCreated)
		if testing.Verbose() {
			t.Logf("Body: %q", rr.Body.Bytes())
		}
	}
	expected = []byte("ca-crt1ca-crt2CA certificatecertificate authorityfake distro CA certs")
	if !bndl.Check() {
		t.Error("bndl.Check() after PUT: got false want true")
	}
	f = bndl.File
	if f == nil {
		t.Fatal("no file metadata after PUT")
	}
	bndlPath = bndl.Path(f)
	if contents, err := ioutil.ReadFile(bndlPath); err != nil {
		t.Fatalf("ioutil.ReadFile(%s) after PUT: %+v", bndlPath, err)
	} else if !bytes.Equal(contents, expected) {
		t.Errorf("file contents after PUT: got %q want %q", contents,
			expected)
	}

	req = httptest.NewRequest(http.MethodPut, reqPath, nil)
	req.Header.Set("Content-Type", "foo/bar")
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusUnsupportedMediaType {
		t.Errorf("PUT status: got %d want %d", rr.Code,
			http.StatusUnsupportedMediaType)
	}
	if rr.Header().Get("Content-Type") != problemContentType {
		t.Errorf("PUT Content-Type: got %s want %s",
			rr.Header().Get("Content-Type"), problemContentType)
	}
	bodylen = len(rr.Body.String())
	if rr.Header().Get("Content-Length") != strconv.Itoa(bodylen) {
		t.Errorf("PUT Content-Length: got %s want %d",
			rr.Header().Get("Content-Length"), bodylen)
	}
	problem = &Problem{}
	if err = json.Unmarshal(rr.Body.Bytes(), problem); err != nil {
		t.Fatalf("PUT body unmarshal: %v", err)
	}
	if problem.Type != errBndlContentType.Type {
		t.Errorf("PUT problem type: got %s want %s", problem.Type,
			errBndlContentType.Type)
	}
	if problem.Title != errBndlContentType.Title {
		t.Errorf("PUT problem title: got %s want %s", problem.Title,
			errBndlContentType.Title)
	}
	errMsg = "Invalid Request Content-Type: foo/bar"
	if problem.Detail != errMsg {
		t.Errorf("PUT problem detail: got \"%s\" want \"%s\"",
			problem.Detail, errMsg)
	}
	if problem.Status != http.StatusUnsupportedMediaType {
		t.Errorf("PUT problem status: got %d want %d", problem.Status,
			http.StatusUnsupportedMediaType)
	}
	if !errInstancePattern.Match([]byte(problem.Instance)) {
		t.Errorf("PUT problem instance: got %s want /log/errors/N",
			problem.Instance)
	}

	req = httptest.NewRequest(http.MethodPut, reqPath, nil)
	req.Header.Set("Content-Type", "application/json")
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusBadRequest {
		t.Errorf("PUT status: got %d want %d", rr.Code,
			http.StatusBadRequest)
	}
	if rr.Header().Get("Content-Type") != problemContentType {
		t.Errorf("PUT Content-Type: got %s want %s",
			rr.Header().Get("Content-Type"), problemContentType)
	}
	bodylen = len(rr.Body.String())
	if rr.Header().Get("Content-Length") != strconv.Itoa(bodylen) {
		t.Errorf("PUT Content-Length: got %s want %d",
			rr.Header().Get("Content-Length"), bodylen)
	}
	problem = &Problem{}
	if err = json.Unmarshal(rr.Body.Bytes(), problem); err != nil {
		t.Fatalf("PUT body unmarshal: %v", err)
	}
	if problem.Type != errBndlBodyUnmarshal.Type {
		t.Errorf("PUT problem type: got %s want %s", problem.Type,
			errBndlBodyUnmarshal.Type)
	}
	if problem.Title != errBndlBodyUnmarshal.Title {
		t.Errorf("PUT problem title: got %s want %s", problem.Title,
			errBndlBodyUnmarshal.Title)
	}
	if len(problem.Detail) == 0 {
		t.Error("PUT problem detail: got empty want non-empty")
	}
	if problem.Status != http.StatusBadRequest {
		t.Errorf("PUT problem status: got %d want %d", problem.Status,
			http.StatusBadRequest)
	}
	if !errInstancePattern.Match([]byte(problem.Instance)) {
		t.Errorf("PUT problem instance: got %s want /log/errors/N",
			problem.Instance)
	}

	srcs = cacrt.BundleSources{
		Srcs: []cacrt.BundleSource{{
			APIVersion: "v4711",
			Kind:       "Secret",
			Namespace:  "ns",
			Name:       "secret",
			Version:    "0815",
			Key:        "crt",
		}},
		Distro: false,
	}
	srcsJSON, err = json.Marshal(srcs)
	if err != nil {
		t.Fatalf("json.Marshal(cacrt.BundleSources): %+v", err)
	}
	body = bytes.NewReader(srcsJSON)
	req = httptest.NewRequest(http.MethodPut, reqPath, body)
	req.Header.Set("Content-Type", "application/json")
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusBadRequest {
		t.Errorf("PUT status: got %d want %d", rr.Code,
			http.StatusBadRequest)
	}
	if rr.Header().Get("Content-Type") != problemContentType {
		t.Errorf("PUT Content-Type: got %s want %s",
			rr.Header().Get("Content-Type"), problemContentType)
	}
	bodylen = len(rr.Body.String())
	if rr.Header().Get("Content-Length") != strconv.Itoa(bodylen) {
		t.Errorf("PUT Content-Length: got %s want %d",
			rr.Header().Get("Content-Length"), bodylen)
	}
	problem = &Problem{}
	if err = json.Unmarshal(rr.Body.Bytes(), problem); err != nil {
		t.Fatalf("PUT body unmarshal: %v", err)
	}
	if problem.Type != errBndlUnsupportedAPIKind.Type {
		t.Errorf("PUT problem type: got %s want %s", problem.Type,
			errBndlUnsupportedAPIKind.Type)
	}
	if problem.Title != errBndlUnsupportedAPIKind.Title {
		t.Errorf("PUT problem title: got %s want %s", problem.Title,
			errBndlUnsupportedAPIKind.Title)
	}
	errMsg = "API/Kind v4711/Secret not supported"
	if problem.Detail != errMsg {
		t.Errorf("PUT problem detail: got \"%s\" want \"%s\"",
			problem.Detail, errMsg)
	}
	if problem.Status != http.StatusBadRequest {
		t.Errorf("PUT problem status: got %d want %d", problem.Status,
			http.StatusBadRequest)
	}
	if !errInstancePattern.Match([]byte(problem.Instance)) {
		t.Errorf("PUT problem instance: got %s want /log/errors/N",
			problem.Instance)
	}

	srcs = cacrt.BundleSources{
		Srcs: []cacrt.BundleSource{{
			APIVersion: "v1",
			Kind:       "Foobar",
			Namespace:  "ns",
			Name:       "foo",
			Version:    "4711",
			Key:        "bar",
		}},
		Distro: false,
	}
	srcsJSON, err = json.Marshal(srcs)
	if err != nil {
		t.Fatalf("json.Marshal(cacrt.BundleSources): %+v", err)
	}
	body = bytes.NewReader(srcsJSON)
	req = httptest.NewRequest(http.MethodPut, reqPath, body)
	req.Header.Set("Content-Type", "application/json")
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusBadRequest {
		t.Errorf("PUT status: got %d want %d", rr.Code,
			http.StatusBadRequest)
	}
	if rr.Header().Get("Content-Type") != problemContentType {
		t.Errorf("PUT Content-Type: got %s want %s",
			rr.Header().Get("Content-Type"), problemContentType)
	}
	bodylen = len(rr.Body.String())
	if rr.Header().Get("Content-Length") != strconv.Itoa(bodylen) {
		t.Errorf("PUT Content-Length: got %s want %d",
			rr.Header().Get("Content-Length"), bodylen)
	}
	problem = &Problem{}
	if err = json.Unmarshal(rr.Body.Bytes(), problem); err != nil {
		t.Fatalf("PUT body unmarshal: %v", err)
	}
	if problem.Type != errBndlUnsupportedAPIKind.Type {
		t.Errorf("PUT problem type: got %s want %s", problem.Type,
			errBndlUnsupportedAPIKind.Type)
	}
	if problem.Title != errBndlUnsupportedAPIKind.Title {
		t.Errorf("PUT problem title: got %s want %s", problem.Title,
			errBndlUnsupportedAPIKind.Title)
	}
	errMsg = "API/Kind v1/Foobar not supported"
	if problem.Detail != errMsg {
		t.Errorf("PUT problem detail: got \"%s\" want \"%s\"",
			problem.Detail, errMsg)
	}
	if problem.Status != http.StatusBadRequest {
		t.Errorf("PUT problem status: got %d want %d", problem.Status,
			http.StatusBadRequest)
	}
	if !errInstancePattern.Match([]byte(problem.Instance)) {
		t.Errorf("PUT problem instance: got %s want /log/errors/N",
			problem.Instance)
	}
}

func TestPutBndlNoDistro(t *testing.T) {
	gid := os.Getgid()
	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	client := fake.NewSimpleClientset()
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)
	bndl, err := cacrt.NewBundle(base, "", gid, getter)
	if err != nil {
		t.Fatalf("NewBundle(): %v", err)
	}

	hndlr := &bndlHndlr{
		log:  &logrus.Logger{Out: ioutil.Discard},
		bndl: bndl,
		crt:  getter,
	}

	srcs := cacrt.BundleSources{
		Distro: true,
	}
	srcsJSON, err := json.Marshal(srcs)
	if err != nil {
		t.Fatalf("json.Marshal(cacrt.BundleSources): %+v", err)
	}
	body := bytes.NewReader(srcsJSON)
	req := httptest.NewRequest(http.MethodPut, reqPath, body)
	req.Header.Set("Content-Type", "application/json")
	rr := httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNotFound {
		t.Errorf("PUT status: got %d want %d", rr.Code,
			http.StatusNotFound)
		if testing.Verbose() {
			t.Logf("Body: %q", rr.Body.Bytes())
		}
	}
	if rr.Header().Get("Content-Type") != problemContentType {
		t.Errorf("PUT Content-Type: got %s want %s",
			rr.Header().Get("Content-Type"), problemContentType)
	}
	bodylen := len(rr.Body.String())
	if rr.Header().Get("Content-Length") != strconv.Itoa(bodylen) {
		t.Errorf("PUT Content-Length: got %s want %d",
			rr.Header().Get("Content-Length"), bodylen)
	}
	problem := &Problem{}
	if err = json.Unmarshal(rr.Body.Bytes(), problem); err != nil {
		t.Fatalf("PUT body unmarshal: %v", err)
	}
	if problem.Type != errBndlNoDistro.Type {
		t.Errorf("PUT problem type: got %s want %s", problem.Type,
			errBndlNoDistro.Type)
	}
	if problem.Title != errBndlNoDistro.Title {
		t.Errorf("PUT problem title: got %s want %s", problem.Title,
			errBndlNoDistro.Title)
	}
	if len(problem.Detail) == 0 {
		t.Error("PUT problem detail: got empty want non-empty")
	}
	errMsg := "No distribution CA certificate bundle configured"
	if !strings.HasPrefix(problem.Detail, errMsg) {
		t.Errorf("PUT problem detail: got \"%s\" want preifx \"%s\"",
			problem.Detail, errMsg)
	}
	if problem.Status != http.StatusNotFound {
		t.Errorf("PUT problem status: got %d want %d", problem.Status,
			http.StatusNotFound)
	}
	if !errInstancePattern.Match([]byte(problem.Instance)) {
		t.Errorf("PUT problem instance: got %s want /log/errors/N",
			problem.Instance)
	}
}

func TestPostBndl(t *testing.T) {
	gid := os.Getgid()
	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	client := fake.NewSimpleClientset(
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "ns1",
				Name:            "secret1",
				UID:             "2572e346-0710-4c3d-a1ae-bacaddb6f3f9",
				ResourceVersion: "4711",
			},
			Data: map[string][]byte{
				"cacrt": []byte("ca-crt1"),
			},
		},
	)
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)
	bndl, err := cacrt.NewBundle(base, "", gid, getter)
	if err != nil {
		t.Fatalf("NewBundle(): %v", err)
	}

	hndlr := &bndlHndlr{
		log:  &logrus.Logger{Out: ioutil.Discard},
		bndl: bndl,
		crt:  getter,
	}

	srcs := cacrt.BundleSources{
		Srcs: []cacrt.BundleSource{{
			APIVersion: "v1",
			Kind:       "Secret",
			Namespace:  "ns1",
			Name:       "secret1",
			Key:        "cacrt",
		}},
		Distro: false,
	}
	srcsJSON, err := json.Marshal(srcs)
	if err != nil {
		t.Fatalf("json.Marshal(cacrt.BundleSources): %+v", err)
	}
	body := bytes.NewReader(srcsJSON)
	req := httptest.NewRequest(http.MethodPost, reqPath, body)
	req.Header.Set("Content-Type", "application/json")
	rr := httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusCreated {
		t.Errorf("POST status: got %d want %d", rr.Code,
			http.StatusCreated)
	}
	expected := []byte("ca-crt1")
	if !bndl.Check() {
		t.Error("bndl.Check() after POST: got false want true")
	}
	f := bndl.File
	if f == nil {
		t.Fatal("no file metadata after POST")
	}
	path := bndl.Path(f)
	if contents, err := ioutil.ReadFile(path); err != nil {
		t.Fatalf("ioutil.ReadFile(%s) after POST: %+v", path, err)
	} else if !bytes.Equal(contents, expected) {
		t.Errorf("file contents after POST: got %q want %q", contents,
			expected)
	}
	if rr.Result().Header.Get("Content-Length") != "0" {
		t.Errorf("POST Content-Length: got %s want 0",
			rr.Result().Header.Get("Content-Length"))
	}

	body.Seek(0, io.SeekStart)
	req = httptest.NewRequest(http.MethodPost, reqPath, body)
	req.Header.Set("Content-Type", "application/json")
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusConflict {
		t.Errorf("2nd POST status: got %d want %d", rr.Code,
			http.StatusConflict)
	}
	if rr.Result().Header.Get("Content-Length") != "0" {
		t.Errorf("2nd POST Content-Length: got %s want 0",
			rr.Result().Header.Get("Content-Length"))
	}
}

func TestDeleteBndl(t *testing.T) {
	gid := os.Getgid()
	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)

	client := fake.NewSimpleClientset(
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "ns1",
				Name:            "secret1",
				UID:             "7cc942f0-22b7-4f83-8282-eba50af4e4a1",
				ResourceVersion: "1234",
			},
			Data: map[string][]byte{
				"cacrt": []byte("ca-crt1"),
			},
		},
		&api_v1.Secret{
			ObjectMeta: metav1.ObjectMeta{
				Namespace:       "ns2",
				Name:            "secret2",
				UID:             "bbf67c5e-b7b1-441d-8f6d-5d3af2cad01d",
				ResourceVersion: "4321",
			},
			Data: map[string][]byte{
				"ca-crt": []byte("ca-crt2"),
			},
		},
	)
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)
	bndl, err := cacrt.NewBundle(base, "", gid, getter)
	if err != nil {
		t.Fatalf("NewBundle(): %v", err)
	}

	file := &cacrt.File{
		Sources: cacrt.BundleSources{
			Srcs: []cacrt.BundleSource{
				{
					APIVersion: "v1",
					Kind:       "Secret",
					Namespace:  "ns1",
					Name:       "secret1",
					UID:        "7cc942f0-22b7-4f83-8282-eba50af4e4a1",
					Version:    "1234",
					Key:        "cacrt",
				},
			},
		},
	}
	if err := bndl.AddOrUpdate(file); err != nil {
		t.Fatalf("bndl.Write(ns1/svc1): %+v", err)
	}

	hndlr := &bndlHndlr{
		log:  &logrus.Logger{Out: ioutil.Discard},
		bndl: bndl,
		crt:  getter,
	}

	req := httptest.NewRequest(http.MethodDelete, reqPath, nil)
	rr := httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNoContent {
		t.Errorf("DELETE status: got %d want %d", rr.Code,
			http.StatusNoContent)
	}
	if bndl.Check() {
		t.Error("bndl.Check() after DELETE: got true want false")
	}
	if rr.Result().Header.Get("Content-Length") != "" {
		t.Errorf("DELETE Content-Length: got %s want empty",
			rr.Result().Header.Get("Content-Length"))
	}

	req = httptest.NewRequest(http.MethodDelete, reqPath, nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNotFound {
		t.Errorf("2nd DELETE status: got %d want %d", rr.Code,
			http.StatusNotFound)
	}
	if rr.Header().Get("Content-Type") != problemContentType {
		t.Errorf("2nd DELETE Content-Type: got %s want %s",
			rr.Header().Get("Content-Type"), problemContentType)
	}
	bodylen := len(rr.Body.String())
	if rr.Header().Get("Content-Length") != strconv.Itoa(bodylen) {
		t.Errorf("2nd DELETE Content-Length: got %s want %d",
			rr.Header().Get("Content-Length"), bodylen)
	}
	problem := &Problem{}
	if err = json.Unmarshal(rr.Body.Bytes(), problem); err != nil {
		t.Fatalf("2nd DELETE body unmarshal: %v", err)
	}
	if problem.Type != errBndlNotFound.Type {
		t.Errorf("2nd DELETE problem type: got %s want %s",
			problem.Type, errBndlNotFound.Type)
	}
	if problem.Title != errBndlNotFound.Title {
		t.Errorf("2nd DELETE problem title: got %s want %s",
			problem.Title, errBndlNotFound.Title)
	}
	if problem.Status != http.StatusNotFound {
		t.Errorf("2nd DELETE problem status: got %d want %d",
			problem.Status, http.StatusNotFound)
	}
	if !errInstancePattern.Match([]byte(problem.Instance)) {
		t.Errorf("2nd DELETE problem instance: "+
			"got %s want /log/errors/N", problem.Instance)
	}

	req = httptest.NewRequest(http.MethodDelete, reqPath, nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNotFound {
		t.Errorf("DELETE status: got %d want %d", rr.Code,
			http.StatusNotFound)
	}

	file = &cacrt.File{
		Sources: cacrt.BundleSources{
			Srcs: []cacrt.BundleSource{
				{
					APIVersion: "v1",
					Kind:       "Secret",
					Namespace:  "ns2",
					Name:       "secret2",
					UID:        "bbf67c5e-b7b1-441d-8f6d-5d3af2cad01d",
					Version:    "4321",
					Key:        "ca-crt",
				},
			},
		},
	}
	if err = bndl.AddOrUpdate(file); err != nil {
		t.Fatalf("bndl.Write(ns2/svc2): %+v", err)
	}
	f := bndl.File
	if f == nil {
		t.Fatal("No File metadata after AddOrUpdate()")
	}
	path := bndl.Path(f)
	if err = os.Remove(path); err != nil {
		t.Fatalf("os.Remove(%s): %+v", path, err)
	}
	req = httptest.NewRequest(http.MethodDelete, reqPath, nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNotFound {
		t.Errorf("DELETE status: got %d want %d", rr.Code,
			http.StatusNotFound)
	}
}

func TestBndl405(t *testing.T) {
	client := fake.NewSimpleClientset()
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)
	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)
	bndl, err := cacrt.NewBundle(base, "", -1, getter)
	if err != nil {
		t.Fatalf("NewBundle(): %v", err)
	}

	hndlr := &bndlHndlr{
		log:  &logrus.Logger{Out: ioutil.Discard},
		bndl: bndl,
		crt:  getter,
	}

	methods := []string{
		http.MethodPatch, http.MethodConnect, http.MethodOptions,
		http.MethodTrace, "FOOBAR",
	}
	for _, method := range methods {
		req := httptest.NewRequest(method, "/v1/ca-bndl/ns/name", nil)
		rr := httptest.NewRecorder()
		hndlr.ServeHTTP(rr, req)
		if rr.Code != http.StatusMethodNotAllowed {
			t.Errorf("%s /v1/ca-bndl/ns/name status: got %d "+
				"want %d", method, rr.Code,
				http.StatusMethodNotAllowed)
		}
		if rr.Result().Header.Get("Allow") != allowRW {
			t.Errorf("%s /v1/ca-bndl/ns/name Allow: got %s want %s",
				method, rr.Result().Header.Get("Allow"),
				allowRW)
		}
		if rr.Result().Header.Get("Content-Length") != "0" {
			t.Errorf("%s /v1/ca-bndl/ns/name Content-Length: "+
				"got %s want 0", method,
				rr.Result().Header.Get("Content-Length"))
		}
	}
}

func TestBndlErrors(t *testing.T) {
	client := fake.NewSimpleClientset()
	lister := setupSecretLister(client)
	getter := crt.NewGetter(lister)
	base, err := getTempDir()
	if err != nil {
		t.Fatalf("ioutil.TempDir(): %+v", err)
	}
	defer os.RemoveAll(base)
	files, err := pem.NewFiles(base, -1, getter)
	if err != nil {
		t.Fatalf("NewFiles(): %v", err)
	}

	hndlr := &pemsHndlr{
		log:       &logrus.Logger{Out: ioutil.Discard},
		files:     files,
		crtGetter: getter,
	}

	req := httptest.NewRequest(http.MethodGet, "/v1/pems/foo", nil)
	rr := httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNotFound {
		t.Errorf("GET /v1/pems/foo status: got %d want %d", rr.Code,
			http.StatusNotFound)
	}
	if rr.Header().Get("Content-Type") != problemContentType {
		t.Errorf("GET /v1/pems/foo Content-Type: got %s want %s",
			rr.Header().Get("Content-Type"), problemContentType)
	}
	bodylen := len(rr.Body.String())
	if rr.Header().Get("Content-Length") != strconv.Itoa(bodylen) {
		t.Errorf("GET /v1/pems/foo Content-Length: got %s want %d",
			rr.Header().Get("Content-Length"), bodylen)
	}
	problem := &Problem{}
	if err = json.Unmarshal(rr.Body.Bytes(), problem); err != nil {
		t.Fatalf("GET /v1/pems/foo body unmarshal: %v", err)
	}
	if problem.Type != errPemsPattern.Type {
		t.Errorf("GET /v1/pems/foo problem type: got %s want %s",
			problem.Type, errPemsPattern.Type)
	}
	if problem.Title != errPemsPattern.Title {
		t.Errorf("GET /v1/pems/foo problem title: got %s want %s",
			problem.Title, errPemsPattern.Title)
	}
	if problem.Detail != errPemsPattern.Detail {
		t.Errorf("GET /v1/pems/foo problem detail: got %s want %s",
			problem.Detail, errPemsPattern.Detail)
	}
	if problem.Status != http.StatusNotFound {
		t.Errorf("GET /v1/pems/foo problem status: got %d want %d",
			problem.Status, http.StatusNotFound)
	}
	if !errInstancePattern.Match([]byte(problem.Instance)) {
		t.Errorf("GET /v1/pems/foo problem instance: "+
			"got %s want /log/errors/N", problem.Instance)
	}
}
