/*
 * Copyright (c) 2020 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package rest

import (
	"io/ioutil"
	"net/http"
	"net/http/httptest"
	"testing"

	"github.com/sirupsen/logrus"
)

func TestHealthz(t *testing.T) {
	version := "47.11"
	hndlr := &healthzHndlr{
		version: version,
		log:     &logrus.Logger{Out: ioutil.Discard},
	}
	rr := httptest.NewRecorder()

	req := httptest.NewRequest(http.MethodGet, "/v1/healthz", nil)
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNoContent {
		t.Errorf("GET /v1/healthz status: got %d want %d",
			rr.Code, http.StatusNoContent)
	}
	if rr.Result().Header.Get("Application-Version") != version {
		t.Errorf("GET /v1/healthz Application-Version: got %s want %s",
			rr.Result().Header.Get("Application-Version"),
			version)
	}
	noStore := "no-store"
	if rr.Result().Header.Get("Cache-Control") != noStore {
		t.Errorf("GET /v1/healthz Cache-Control: got %s want %s",
			rr.Result().Header.Get("Cache-Control"),
			noStore)
	}
	if rr.Result().Header.Get("Content-Length") != "" {
		t.Errorf("GET /v1/healthz Content-Length: got %s want empty",
			rr.Result().Header.Get("Content-Length"))
	}

	req = httptest.NewRequest(http.MethodHead, "/v1/healthz", nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNoContent {
		t.Errorf("HEAD /v1/healthz status: got %d want %d",
			rr.Code, http.StatusNoContent)
	}

	req = httptest.NewRequest(http.MethodGet, "/v1/healthz/foo", nil)
	rr = httptest.NewRecorder()
	hndlr.ServeHTTP(rr, req)
	if rr.Code != http.StatusNotFound {
		t.Errorf("GET /v1/healthz/foo status: got %d want %d",
			rr.Code, http.StatusNotFound)
	}
	if rr.Result().Header.Get("Content-Length") != "0" {
		t.Errorf("GET /v1/healthz/foo Content-Length: got %s want 0",
			rr.Result().Header.Get("Content-Length"))
	}
}

func TestHealthz405(t *testing.T) {
	version := "08.15"
	hndlr := &healthzHndlr{
		version: version,
		log:     &logrus.Logger{Out: ioutil.Discard},
	}
	noStore := "no-store"

	methods := []string{
		http.MethodPost, http.MethodPut, http.MethodPatch,
		http.MethodDelete, http.MethodConnect, http.MethodOptions,
		http.MethodTrace, "FOOBAR",
	}
	for _, method := range methods {
		req := httptest.NewRequest(method, "/v1/healthz", nil)
		rr := httptest.NewRecorder()
		hndlr.ServeHTTP(rr, req)
		if rr.Code != http.StatusMethodNotAllowed {
			t.Errorf("%s /v1/healthz status: got %d want %d",
				method, rr.Code, http.StatusMethodNotAllowed)
		}
		if rr.Result().Header.Get("Allow") != allowRdonly {
			t.Errorf("%s /v1/healthz Allow: got %s want %s",
				method, rr.Result().Header.Get("Allow"),
				allowRdonly)
		}
		if rr.Result().Header.Get("Content-Length") != "0" {
			t.Errorf("%s /v1/healthz Content-Length: got %s want 0",
				method, rr.Result().Header.
					Get("Content-Length"))
		}
		if rr.Result().Header.Get("Cache-Control") != noStore {
			t.Errorf("%s /v1/healthz Cache-Control: got %s want %s",
				method, rr.Result().Header.Get("Cache-Control"),
				noStore)
		}
	}
}
